const { plugin: { pluginScope } } = require('@kl/constants');
const lo = require('lodash');
const hlp = require('../../helpers');
const hlpPolicy = require('../../helpers_policy');
const consts = require('../../consts');
const apiUtils = require('../../common/open_api');
const hostedDef = require('../../default/policy-hosted.js');
const policyConverters = require('./sections/index.js');

function Policy () {
    let policiesCache = {policyId: -1, changed:[]};

    async function CacheFromTempDataStore(tempDataStore) {
        if(tempDataStore!==undefined) {
            policiesCache = await tempDataStore.get('policies_cache');
            if(policiesCache===undefined || policiesCache==={}){
                policiesCache= {policyId: -1, changed:[]};
            }
        }
    }

    async function CacheToTempDataStore(tempDataStore) {
        if(tempDataStore!==undefined) {
            await tempDataStore.set('policies_cache', policiesCache);
        }
    }

    function AddToChangesList (arrSections) {
        if(arrSections===undefined)
            return;
        arrSections = Array.isArray(arrSections) ? arrSections : [arrSections];
        for (let policySection of arrSections) {
            if (policySection!=='' && !policiesCache.changed.includes(policySection)) {
                policiesCache.changed.push(policySection);
            }
        }
    }

    function UpdatePoliciesCacheAsDefault (arrSections, cache) {
        arrSections = Array.isArray(arrSections) ? arrSections : [arrSections];
        for (let policySection of arrSections) {
            if (policySection!=='' && cache[policySection]===undefined ) {
                cache[policySection] = hlpPolicy.DefaultPolicyObject(policySection);
                runtime.logger.log('KSWS: Cached Default Policy: ' + policySection + ' Data:' + JSON.stringify(cache[policySection]));
            }
        }
        return cache;
    }

    async function UpdateRbac (api, id, rowData, sectionId) {
        let policyData = hlp.serialize(rowData);
        try {
            const readOnlyNodes = await api.HstAccessControl.GetPolicyReadonlyNodes({
                IGroupId: id,
                szwProduct: consts.wstrProduct,
                szwVersion: consts.wstrVersion,
                szwSectionName: sectionId,
                pPolicySection: policyData.Policy!==undefined ? policyData.Policy.value : policyData,
            });

            if (readOnlyNodes.result!==undefined ) {
                let changingData = rowData.Policy!==undefined ? rowData.Policy.KLPRSS_Val : rowData;
                hlpPolicy.SetPolicyFieldsRbacLocked(changingData, readOnlyNodes.result[sectionId]);
            }
        }
        catch (error) {
            runtime.logger.error('KSWS Error: ' + error);
        }

        return rowData;
    }
    /*
     * Метод добавляет в cache политик новые секции
     */
    async function UpdatePoliciesCache (api, id, arrSections, parentId) {
        arrSections = Array.isArray(arrSections) ? arrSections : [arrSections];

        if (id===undefined || id === -1) {
            policiesCache = UpdatePoliciesCacheAsDefault(arrSections, policiesCache);
            return;
        }

        let { result: identifier } = await api.Policy.GetPolicyContents({ nPolicy: id });

        for (let i = 0; i < arrSections.length; i++) {
            const policySection = arrSections[i];
            if( policiesCache[policySection] !== undefined ){
                continue;
            }

            const SsPolicySection = consts.policySsSections[policySection];
            if(SsPolicySection === undefined)
                continue;

            let sectionData = await api.SsContents.Ss_Read({
                wstrID: identifier,
                wstrProduct: consts.wstrProduct,
                wstrVersion: consts.wstrVersion,
                wstrSection: SsPolicySection,
                skipDeserialization: true});

            let error = lo.get(sectionData, 'PxgError');
            if (error) {
                policiesCache[policySection] = hlpPolicy.DefaultPolicyObject(policySection);
            }
            else {
                let rowData = (sectionData=== undefined) ? hlpPolicy.DefaultPolicyObject(policySection) : hlp.deserialize(lo.get(sectionData, 'result')).KLPRSS_Val;
                rowData = await UpdateRbac (api, parentId, rowData, SsPolicySection);
                policiesCache[policySection] = rowData;
                runtime.logger.log('KSWS: Cached Policy: ' + policySection + ' Data:' + JSON.stringify(rowData));
            }
        }
        api.SsContents.Ss_Release({ wstrID: identifier });
    }

    async function SaveDefaultPolicySections (api, id) {
        if( id === -1  || id === undefined)
            return;

        const bIsHosted = await apiUtils.IsSetHostedMode(api);

        let { result: identifier } = await api.Policy.GetPolicyContents({ nPolicy: id });
        for (let keySection in consts.policySsSections) {
            if(!hostedDef.needToCreatePolicySection(bIsHosted, keySection))
                continue;

            const bChangedData = ( policiesCache.changed.includes(keySection) );
            let DataToSave = {
                KLPRSS_Val: bChangedData ? policiesCache[keySection] : hlpPolicy.DefaultPolicyObject(keySection)
            };

            if( DataToSave.KLPRSS_Val.length === 0 )
                continue;
            if(bIsHosted && keySection === 'NetworkLists'){
                DataToSave.KLPRSS_Val = hostedDef.defaultHostedNetworkLists();
            }

            const SsPolicySection = consts.policySsSections[keySection];
            let result = hlp.serialize(DataToSave);
            result.KLPRSS_Mnd = false;

            runtime.logger.log('KSWS: Save default Section: ' + SsPolicySection);
            runtime.logger.log('KSWS: Data: ' + JSON.stringify(result));

            await api.SsContents.Ss_Replace({
                wstrID: identifier,
                wstrProduct: consts.wstrProduct,
                wstrVersion: consts.wstrVersion,
                wstrSection: SsPolicySection,
                pNewData: result,
                skipSerialization: true
            });
            await api.SsContents.Ss_Apply({ wstrID: identifier });
        }
        api.SsContents.Ss_Release({ wstrID: identifier });
    }

    async function SaveChangedPolicySections (api, id) {
        if(policiesCache.changed.length === 0 || id === -1  || id === undefined)
            return;

        let { result: identifier } = await api.Policy.GetPolicyContents({ nPolicy: id });
        for (let policySection of policiesCache.changed) {
            const SsPolicySection = consts.policySsSections[policySection];
            let DataToSave = { KLPRSS_Val: policiesCache[policySection] };
            let result = hlp.serialize(DataToSave);
            result.KLPRSS_Mnd = false;

            await api.SsContents.Ss_Replace({
                wstrID: identifier,
                wstrProduct: consts.wstrProduct,
                wstrVersion: consts.wstrVersion,
                wstrSection: SsPolicySection,
                pNewData: result,
                skipSerialization: true
            });
            await api.SsContents.Ss_Apply({ wstrID: identifier });
        }
        api.SsContents.Ss_Release({ wstrID: identifier });
    }

    /*
     * Called via api.js in getFormModel
     */
    async function getPolicyData (api, id, parentId, section, tDataStore) {
        if(policiesCache.policyId !== id) {
            policiesCache = {policyId:id, changed:[]};
            await CacheToTempDataStore(tDataStore);
        }

        if (section === undefined || policyConverters[section] === undefined)
            return {};

        await CacheFromTempDataStore(tDataStore);
        if (policyConverters[section].sections !== undefined) {
            await UpdatePoliciesCache(api, id, policyConverters[section].sections, parentId);
        }
        await CacheToTempDataStore(tDataStore);

        if (policyConverters[section].toViewEx !== undefined) {
            try {
                return await policyConverters[section].toViewEx(api, policiesCache);
            } catch(error){
                runtime.logger.error('KSWS toViewEx: ' + error);
                return {};
            }
        }

        if (policyConverters[section].toView !== undefined) {
            try {
                const computedData = policyConverters[section].toView(policiesCache);
                runtime.logger.log('KSWS: getPolicyData. Data: ' + JSON.stringify(computedData));
                return computedData;
            } catch(error){
                runtime.logger.error('KSWS toView: ' + error);
                return {};
            }
        }

        return {};
    }
    /*
     * Called via api.js in setFormModel
     */
    async function updatePolicyData (kscApi, section, data, tDataStore) {
        if (section === undefined || policyConverters[section] === undefined)
            return;

        await CacheFromTempDataStore(tDataStore);
        AddToChangesList(policyConverters[section].sections);
        if(policyConverters[section].toRawEx!==undefined)
            await policyConverters[section].toRawEx(kscApi, data, policiesCache);
        else
            policyConverters[section].toRaw(data, policiesCache);
        await CacheToTempDataStore(tDataStore);
    }
    /*
     * Called via api.js in beforeApplyChanges
     */
    function commitPolicyData (api, scope) {
    }
    /*
     * Called via api.js in applyChanges
     */
    async function savePolicyData (api, id, scope, tDataStore) {
        await CacheFromTempDataStore(tDataStore);

        if (scope === pluginScope.newObject || scope === pluginScope.newQSWObject || scope === pluginScope.newCloudProtectionObject) {
            runtime.logger.log('KSWS: SaveDefaultPolicySections');
            await SaveDefaultPolicySections (api, id);
        }
        else{
            await SaveChangedPolicySections(api, id);
        }

        policiesCache = {policyId: -1, changed:[]};
        await CacheToTempDataStore(tDataStore);
    }

    return {
        getPolicyData,
        updatePolicyData,
        commitPolicyData,
        savePolicyData
    }
}

module.exports = Policy;
